"""Utilities for creating Sankey plots of pystv results."""
from dataclasses import dataclass, field

import plotly.express as px
import plotly.graph_objects as go

NODE_PALETTE = px.colors.qualitative.Dark2
LINK_PALETTE = px.colors.qualitative.Set2
NUM_COLORS = len(NODE_PALETTE)


@dataclass
class SankeyData:
    """Container for data useful for creating Sankey Plots."""

    source: list = field(default_factory=list)
    target: list = field(default_factory=list)
    value: list = field(default_factory=list)
    link_color: list = field(default_factory=list)

    labels: list = field(default_factory=list)
    node_color: list = field(default_factory=list)


def result_to_sankey_data(race_result):
    num_rounds = len(race_result.rounds)
    num_cands = len(race_result.rounds[0].count)

    node_color = [NODE_PALETTE[idx % NUM_COLORS] for idx in range(num_cands)]

    data = SankeyData()
    data.node_color = node_color * num_rounds
    data.labels = race_result.metadata.names * num_rounds

    for rnd in range(num_rounds - 1):
        round_result = race_result.rounds[rnd]
        offset = rnd * num_cands
        for src in range(num_cands):
            src_left = round_result.count[src]
            link = LINK_PALETTE[src % NUM_COLORS]
            if src in round_result.transfers:
                for tgt, diff in round_result.transfers[src].items():
                    data.source.append(src + offset)
                    data.target.append(tgt + offset + num_cands)
                    data.value.append(diff)
                    data.link_color.append(link)
                    src_left -= diff
            if src_left > 0:
                data.source.append(src + offset)
                data.target.append(src + offset + num_cands)
                data.value.append(src_left)
                data.link_color.append(link)

    return data


def create_sankey_fig(sankey_data):
    sankey = go.Sankey(
        node=dict(
            thickness=10,
            line=dict(color="black", width=1),
            label=sankey_data.labels,
            color=sankey_data.node_color,
        ),
        link=dict(
            source=sankey_data.source,
            target=sankey_data.target,
            value=sankey_data.value,
            color=sankey_data.link_color,
        ),
        visible=True,
    )

    fig = go.Figure(data=[sankey])
    notes = [
        "Height of bars and traces is proportional to vote count.",
    ]
    fig.add_annotation(
        x=0,
        y=-0.2,
        xref="paper",
        yref="paper",
        text="<br>".join(notes),
        align="left",
        showarrow=False,
        font_size=15,
    )
    fig.update_layout(margin=dict(l=0, t=0))
    return fig
